import abc
import unittest
import numpy as np
import six

from scipy import stats

from kafe2.core.minimizers import AVAILABLE_MINIMIZERS

from kafe2.config import kc

from kafe2.fit import UnbinnedFit, UnbinnedContainer
from kafe2.fit.unbinned.cost import UnbinnedCostFunction_NegLogLikelihood

from kafe2.test.fit.test_fit import AbstractTestFit



def unbinned_model_density(x, tau=2.2, fbg=0.1):
    b = 11.5
    a = 1.
    pdf1 = np.exp(-x / tau) / tau / (np.exp(-a / tau) - np.exp(-b / tau))
    pdf2 = 1. / (b - a)
    return (1 - fbg) * pdf1 + fbg * pdf2


class TestUnbinnedFitBasicInterface(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):

        self._ref_data = np.array([
            7.42, 3.773, 5.968, 4.924, 1.468, 4.664, 1.745, 2.144, 3.836, 3.132, 1.568, 2.352,
            2.132, 9.381, 1.484, 1.181, 5.004, 3.06, 4.582, 2.076, 1.88, 1.337, 3.092, 2.265,
            1.208, 2.753, 4.457, 3.499, 8.192, 5.101, 1.572, 5.152, 4.181, 3.52, 1.344, 10.29,
            1.152, 2.348, 2.228, 2.172, 7.448, 1.108, 4.344, 2.042, 5.088, 1.02, 1.051, 1.987,
            1.935, 3.773, 4.092, 1.628, 1.688, 4.502, 4.687, 6.755, 2.56, 1.208, 2.649, 1.012,
            1.73, 2.164, 1.728, 4.646, 2.916, 1.101, 2.54, 1.02, 1.176, 4.716, 9.671, 1.692,
            9.292, 10.72, 2.164, 2.084, 2.616, 1.584, 5.236, 3.663, 3.624, 1.051, 1.544, 1.496,
            1.883, 1.92, 5.968, 5.89, 2.896, 2.76, 1.475, 2.644, 3.6, 5.324, 8.361, 3.052, 7.703,
            3.83, 1.444, 1.343])

        self._ref_cont = UnbinnedContainer(self._ref_data)

        # reference initial values
        self._ref_initial_pars = np.array([2.2, 0.1])
        self._ref_initial_model = unbinned_model_density(self._ref_data, *self._ref_initial_pars)

        # pre-fit cost value
        self._ref_initial_cost = -2*np.sum(np.log(self._ref_initial_model))

        # reference fit result values
        self._nominal_fit_result_pars = np.array([2.12812181, 0.11117378])
        self._nominal_fit_result_model = unbinned_model_density(self._ref_data, *self._nominal_fit_result_pars)

        self._nominal_fit_result_cost = -2*np.sum(np.log(self._nominal_fit_result_model))

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=1,

            parameter_values=self._ref_initial_pars,
            parameter_names=('tau', 'fbg'),
            cost_function_value=self._ref_initial_cost,

            data=self._ref_data,
            model=self._ref_initial_model,
        )

    def _get_fit(self, model_density_function=None, cost_function=None):
        '''convenience'''

        model_density_function = model_density_function or unbinned_model_density

        # TODO: fix default
        cost_function = cost_function or UnbinnedCostFunction_NegLogLikelihood()

        _fit = UnbinnedFit(
            data=self._ref_cont,
            model_function=model_density_function,
            cost_function=cost_function,
            minimizer=self.MINIMIZER
        )

        return _fit

    def _get_test_fits(self):
        return {
            'default': \
                self._get_fit(),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(
                self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                model=self._nominal_fit_result_model,
                did_fit=True,
                cost_function_value=np.float64(self._nominal_fit_result_cost),
            ),
            fit_names=['default'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_update_data(self):
        _fit = self._get_fit()

        _new_data = np.array([
            1.891075005713712, 2.361504517508682, 1.2026086768765802, 2.0192327689754013,
            6.789827691034451, 6.229060175471942, 5.6827777380476725, 1.3896314892970838,
            1.0379155165227294, 1.1948753489234705, 1.8735602750435354, 2.139305908923123,
            1.935777036301727, 6.108651728684855, 1.5185867568257783, 1.1244705951426477,
            2.2175887045709146, 3.900030680803669, 4.110502050796017, 3.634194797012612,
            4.931709433880977, 3.779552162832593, 3.6401901478360594, 1.7084479392384508,
            10.502957316458524, 1.669761159672512, 2.0378350020970446, 1.4406236513060318,
            2.3226997026175917, 1.2530908142355945, 2.4681349735479428, 2.2697496807176245,
            8.056140162306527, 6.483413760658378, 2.0103631281648733, 1.8484180369382175,
            2.042048282225629, 1.826991555740888, 4.119685425461164, 2.8467560883558485,
            1.8277818983512852, 9.894627016335363, 1.0841480000087294, 3.1732616953846584,
            1.6776961308198861, 1.555131926310589, 1.1806640179159833, 3.052979898534809,
            5.059864773115917, 1.1294288761955027, 2.6979674092930193, 9.82656489928625,
            5.501267559105623, 1.2351970327509707, 1.6078922552697488, 1.582561168396548,
            2.391340419810505, 2.660167843760268, 1.9569767827507456, 1.2776674104800188,
            1.6183122532590568, 4.552555359280379, 4.214077308248084, 1.7815460978157334,
            3.5980732014726056, 1.690396198372471, 4.671077054310741, 1.2339939189715687,
            2.7618415144011874, 2.86744016658739])

        _fit.data = UnbinnedContainer(_new_data)

        self._assert_fit_properties(
            _fit,
            dict(
                data=_new_data,
            )
        )

        _fit.do_fit()

        self._assert_fit_properties(
            _fit,
            dict(
                data=_new_data,
                parameter_values=np.array([1.64489992, 0.1225065]),
            ),
            rtol=1e-2
        )

    def test_reserved_parameter_names_raise(self):
        def dummy_model(x, data):
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('reserved', _exc.exception.args[0])
        self.assertIn('data', _exc.exception.args[0])

    def test_model_no_pars_raise(self):
        def dummy_model():
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn(
            'needs at least one parameter',
            _exc.exception.args[0])

    def test_model_no_pars_beside_x_raise(self):
        def dummy_model(x):
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn(
            'needs at least one parameter besides',
            _exc.exception.args[0])

    def test_model_varargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])

    def test_model_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varkwargs', _exc.exception.args[0])

    def test_model_varargs_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            UnbinnedFit(data=self._ref_cont,
                    model_function=dummy_model,
                    minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])
        # TODO: enable when implemented
        #self.assertIn('varkwargs', _exc.exception.args[0])

    def test_report_before_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_report_after_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.do_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_fit_scalar_model_function(self):
        def _model_function(x, a):
            return a

        _fit = UnbinnedFit([1.0, 2.0], _model_function)
        _fit.limit_parameter("a", 0.001, 0.999)
        _fit.do_fit()
        self.assertAlmostEqual(_fit.parameter_values[0], 0.999)
        _fit.report()



import numpy  # help IDEs with type-hinting inside docstrings

from .._base import PlotAdapterBase
from .._aux import add_pad_to_range
np = numpy


__all__ = ["XYPlotAdapter"]


class XYPlotAdapter(PlotAdapterBase):

    PLOT_STYLE_CONFIG_DATA_TYPE = 'xy'
    PLOT_SUBPLOT_TYPES = dict(
        PlotAdapterBase.PLOT_SUBPLOT_TYPES,
        model_line=dict(
            plot_adapter_method='plot_model_line',
            target_axes='main'
        ),
        model_error_band=dict(
            plot_adapter_method='plot_model_error_band',
            target_axes='main'
        ),
        ratio_error_band=dict(
            plot_style_as='model_error_band',
            plot_adapter_method='plot_ratio_error_band',
            target_axes='ratio'
        ),
        residual_error_band=dict(
            plot_style_as='model_error_band',
            plot_adapter_method='plot_residual_error_band',
            target_axes='residual'
        ),
    )

    AVAILABLE_X_SCALES = ('linear', 'log')

    def __init__(self, xy_fit_object, from_container=False):
        """Construct an :py:obj:`XYPlotContainer` for a :py:obj:`~.XYFit` object:

        :param kafe2.XYFit xy_fit_object: The :py:obj:`~.XYFit` object handled by this plot
            adapter.
        :param from_container: Whether xy_fit_object was created ad-hoc from just a data container.
        :type from_container: bool
        """
        self._fit = xy_fit_object  # needed for type hinting to work correctly
        super(XYPlotAdapter, self).__init__(fit_object=xy_fit_object, from_container=from_container)
        self.n_plot_points = max(200, 4*len(self.data_x))
        self.x_range = add_pad_to_range(self._fit.x_range, scale=self.x_scale)

    # -- public properties

    @property
    def data_x(self):
        return self._fit.x_data

    @property
    def data_y(self):
        return self._fit.y_data

    @property
    def data_xerr(self):
        return self._fit.x_total_error

    @property
    def data_yerr(self):
        return self._fit.y_total_error

    @property
    def model_x(self):
        return self._fit.x_model

    @property
    def model_y(self):
        return self._fit.y_model

    @property
    def model_xerr(self):
        return self._fit.x_model_error

    @property
    def model_yerr(self):
        return self._fit.y_model_error

    @PlotAdapterBase.x_scale.setter
    def x_scale(self, scale):
        update_xrange = self.x_range == add_pad_to_range(self._fit.x_range, scale=self.x_scale)
        PlotAdapterBase.x_scale.fset(self, scale)  # use parent setter
        if update_xrange:
            self.x_range = add_pad_to_range(self._fit.x_range, scale=self.x_scale)

    @property
    def model_line_x(self):
        """*x* support values for model function. Adapts spacing to :py:obj:`.x_scale`.

        :rtype: numpy.ndarray[float]
        """
        _xmin, _xmax = self.x_range
        if self.x_scale == 'linear':
            return np.linspace(_xmin, _xmax, self.n_plot_points)
        if self.x_scale == 'log':
            try:
                return np.geomspace(_xmin, _xmax, self.n_plot_points)
            except ValueError as _e:
                raise ValueError("Support point calculation failed. "
                                 "The plot range can't include 0 when using log scale.") from _e
        raise ValueError("x_range has to be one of {}. Found {} instead.".format(
            self.AVAILABLE_X_SCALES, self.x_scale))

    @property
    def model_line_y(self):
        """*y* values of the model function at the support points :py:obj:`.model_line_x`.

        :rtype: numpy.ndarray[float]
        """
        return self._fit.eval_model_function(x=self.model_line_x)

    @property
    def y_error_band(self):
        """1D array representing the uncertainty band around the model function at the support
        points :py:obj:`.model_line_x`.

        :rtype: numpy.ndarray[float]
        """
        return self._fit.error_band(self.model_line_x)

    # public methods

    def plot_data(self, target_axes, error_contributions=('data',), **kwargs):
        """Plot the measurement data to a specified :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param error_contributions: Which error contributions to include when plotting the data.
            Can either be ``data``, ``'model'`` or both.
        :type error_contributions: str or Tuple[str]
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.errorbar`.
        :return: plot handle(s)
        """

        _yerr = self._get_total_error(error_contributions)

        return target_axes.errorbar(self.data_x,
                                    self.data_y,
                                    xerr=self.data_xerr,
                                    yerr=_yerr,
                                    **kwargs)

    def plot_model(self, target_axes, error_contributions=('model',), **kwargs):
        """Plot the model data to a specified :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param error_contributions: Which error contributions to include when plotting the model.
        :type error_contributions: str or Tuple[str]
            Can either be ``data``, ``'model'`` or both.
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.errorbar`.
        :return: plot handle(s)
        """

        _yerr = self._get_total_error(error_contributions)

        return target_axes.errorbar(self.model_x,
                                    self.model_y,
                                    xerr=self.data_xerr,
                                    yerr=_yerr,
                                    **kwargs)

    def plot_model_line(self, target_axes, **kwargs):
        """Plot the model function to a specified :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.plot`.
        :return: plot handle(s)
        """
        # TODO: how to handle 'data' errors and 'model' errors?
        return target_axes.plot(self.model_line_x,
                                self.model_line_y,
                                **kwargs)

    def plot_model_error_band(self, target_axes, **kwargs):
        """Plot an error band around the model model function.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.fill_between`.
        :return: plot handle(s)
        """
        if self._fit.errors_valid:
            _band_y = self.y_error_band
            _y = self.model_line_y
            return target_axes.fill_between(
                self.model_line_x,
                _y - _band_y, _y + _band_y,
                **kwargs)
        return None  # don't plot error band if fitter input data has no errors...

    def plot_ratio_error_band(self, target_axes, **kwargs):
        """Plot model error band around the data/model ratio to specified
        :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.fill_between`.
        :return: plot handle(s)
        """
        if self._fit.did_fit and (
                self._fit.has_errors or not self._fit._cost_function.needs_errors):
            _band_y = self.y_error_band
            _y = self.model_line_y
            return target_axes.fill_between(
                self.model_line_x,
                1 - _band_y/_y, 1 + _band_y/_y,
                **kwargs)
        return None  # don't plot error band if fitter input data has no errors...

    def plot_residual_error_band(self, target_axes, **kwargs):
        """Plot model error band around the data/model ratio to specified
        :py:obj:`matplotlib.axes.Axes` object.

        :param matplotlib.axes.Axes target_axes: The :py:obj:`matplotlib` axes used for plotting.
        :param dict kwargs: Keyword arguments accepted by :py:obj:`matplotlib.pyplot.fill_between`.
        :return: plot handle(s)
        """
        if self._fit.did_fit and (
                self._fit.has_errors or not self._fit._cost_function.needs_errors):
            _band_y = self.y_error_band
            return target_axes.fill_between(
                self.model_line_x,
                -_band_y, _band_y,
                **kwargs)
        return None  # don't plot error band if fitter input data has no errors...

    def update_plot_kwargs(self, plot_type, plot_kwargs):
        # update ratio kwargs as well, when corresponding plot_types are updated
        # can be overwritten by the user by explicitly setting the ratio kwargs last
        if plot_type == 'data':
            super(XYPlotAdapter, self).update_plot_kwargs(plot_type='ratio', plot_kwargs=plot_kwargs)
        elif plot_type == 'model_error_band':
            super(XYPlotAdapter, self).update_plot_kwargs(plot_type='ratio_error_band', plot_kwargs=plot_kwargs)
        super(XYPlotAdapter, self).update_plot_kwargs(plot_type=plot_type, plot_kwargs=plot_kwargs)



import unittest
import numpy as np
from six import StringIO

from kafe2.fit._base import ModelFunctionFormatter, ModelFunctionBase
from kafe2.fit.histogram import HistModelFunction
from kafe2.fit.indexed import IndexedModelFunction, IndexedModelFunctionFormatter
from kafe2.fit.representation import ModelFunctionYamlWriter, ModelFunctionYamlReader
from kafe2.fit.io.handle import IOStreamHandle

TEST_MODEL_FUNCTION_HIST = """
type: histogram
python_code: 'x a b -> a * x + b'
"""

TEST_MODEL_FUNCTION_HIST_MISSING_KEYWORD="""
type: histogram
"""

TEST_MODEL_FUNCTION_HIST_EXTRA_KEYWORD = TEST_MODEL_FUNCTION_HIST + """
extra_keyword: 3.14
"""

TEST_MODEL_FUNCTION_HIST_WITH_FORMATTER = TEST_MODEL_FUNCTION_HIST + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    arg_formatters:
      x: r
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {x} + {1}'
    latex_expression_string: '{0}{x} + {1}' 
"""

class TestHistModelFunctionYamlRepresenter(unittest.TestCase):

    @staticmethod
    def linear_model(x, a, b):
        return a * x + b
    
    def setUp(self):
        self._test_x = np.arange(10)
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_x, self._test_a, self._test_b)
        
        self._model_function = HistModelFunction(self.linear_model)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_WITH_FORMATTER))
        
        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_HIST_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_x, self._test_a, self._test_b),
                self._test_y
            )
        )

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_x, self._test_a, self._test_b),
                self._test_y
            )
        )
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == 'linear_model')
        self.assertTrue(_read_formatter.latex_name == 'linear model')
        self.assertTrue(_read_arg_formatters[0].name == 'x')
        self.assertTrue(_read_arg_formatters[0].latex_name == r'{r}')
        self.assertTrue(_read_arg_formatters[1].name == 'alpha')
        self.assertTrue(_read_arg_formatters[1].latex_name == r'{\alpha}')
        self.assertTrue(_read_arg_formatters[2].name == 'beta')
        self.assertTrue(_read_arg_formatters[2].latex_name == r'{\beta}')
        self.assertTrue(_read_formatter.expression_format_string == '{0} * {x} + {1}')
        self.assertTrue(_read_formatter.latex_expression_format_string == '{0}{x} + {1}')

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, HistModelFunction))

        self.assertTrue(
            np.allclose(
                self._test_y,
                _read_model_function.func(self._test_x, self._test_a, self._test_b)
            )
        )
        
        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter
        
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        
        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters
        
        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_arg_formatters[2].name == _given_arg_formatters[2].name)
        self.assertTrue(_read_arg_formatters[2].latex_name == _given_arg_formatters[2].latex_name)
        self.assertTrue(_read_formatter.expression_format_string ==  _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string ==  _given_formatter.latex_expression_format_string)


TEST_MODEL_FUNCTION_INDEXED = """
type: indexed
python_code: |
    def linear_model(a, b):
        return a * np.arange(10) + b
"""

TEST_MODEL_FUNCTION_INDEXED_MISSING_KEYWORD = """
type: indexed
"""

TEST_MODEL_FUNCTION_INDEXED_EXTRA_KEYWORD = TEST_MODEL_FUNCTION_INDEXED + """
extra_keyword: 3.14
"""

TEST_MODEL_FUNCTION_INDEXED_WITH_FORMATTER = TEST_MODEL_FUNCTION_INDEXED + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    index_name: r
    latex_index_name: r
    arg_formatters:
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {r} + {1}'
    latex_expression_string: '{0}{r} + {1}' 
"""


class TestIndexedModelFunctionYamlRepresenter(unittest.TestCase):

    TEST_X = np.arange(10)

    @staticmethod
    def linear_model(a, b):
        #TODO handle
        #return a * TestIndexedModelFunctionYamlRepresenter.TEST_X + b
        return a * np.arange(10) + b
    
    def setUp(self):
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_a, self._test_b)
        
        self._model_function = IndexedModelFunction(self.linear_model)
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_WITH_FORMATTER))
        
        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_a, self._test_b),
                self._test_y
            )
        )

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_a, self._test_b),
                self._test_y
            )
        )
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, IndexedModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == 'linear_model')
        self.assertTrue(_read_formatter.latex_name == 'linear model')
        self.assertTrue(_read_formatter.index_name == 'r')
        self.assertTrue(_read_formatter.latex_index_name == r'{r}')
        self.assertTrue(_read_arg_formatters[0].name == 'alpha')
        self.assertTrue(_read_arg_formatters[0].latex_name == r'{\alpha}')
        self.assertTrue(_read_arg_formatters[1].name == 'beta')
        self.assertTrue(_read_arg_formatters[1].latex_name == r'{\beta}')
        self.assertTrue(_read_formatter.expression_format_string == '{0} * {r} + {1}')
        self.assertTrue(_read_formatter.latex_expression_format_string == '{0}{r} + {1}')

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, IndexedModelFunction))

        self.assertTrue(
            np.allclose(
                self._test_y,
                _read_model_function.func(self._test_a, self._test_b)
            )
        )
        
        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter
        
        self.assertTrue(isinstance(_read_formatter, IndexedModelFunctionFormatter))
        
        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters
        
        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_formatter.index_name == _given_formatter.index_name)
        self.assertTrue(_read_formatter.latex_index_name == _given_formatter.latex_index_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_formatter.expression_format_string ==  _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string ==  _given_formatter.latex_expression_format_string)


TEST_MODEL_FUNCTION_BASE = """
python_code: 'x a b -> a * x + b'
"""

TEST_MODEL_FUNCTION_BASE_MISSING_KEYWORD = """
type: base
"""

TEST_MODEL_FUNCTION_XY_EXTRA_KEYWORD = TEST_MODEL_FUNCTION_BASE + """
extra_keyword: 3.14
"""

TEST_MODEL_FUNCTION_XY_WITH_FORMATTER = TEST_MODEL_FUNCTION_BASE + r"""
model_function_formatter:
    name: linear_model
    latex_name: linear model
    arg_formatters:
      x: r
      a:
        name: alpha
        latex_name: \alpha
      b:
        name: beta
        latex_name: \beta
    expression_string: '{0} * {x} + {1}'
    latex_expression_string: '{0}{x} + {1}' 
"""


class TestModelFunctionBaseYamlRepresenter(unittest.TestCase):

    @staticmethod
    def linear_model(x, a, b):
        return a * x + b
    
    def setUp(self):
        self._test_x = np.arange(10)
        self._test_a = 2.0
        self._test_b = -1.0
        self._test_y = self.linear_model(self._test_x, self._test_a, self._test_b)
        
        self._model_function = ModelFunctionBase(self.linear_model)
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_BASE))
        self._testfile_stringstream_with_formatter = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_XY_WITH_FORMATTER))
        
        self._roundtrip_streamreader = ModelFunctionYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionYamlWriter(self._model_function, self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionYamlReader(self._testfile_stringstream)
        self._testfile_streamreader_with_formatter = ModelFunctionYamlReader(self._testfile_stringstream_with_formatter)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_BASE_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_XY_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionYamlReader(self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_x, self._test_a, self._test_b),
                self._test_y
            )
        )

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_read_from_testfile_stream_with_formatter(self):
        _read_model_function = self._testfile_streamreader_with_formatter.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))
        self.assertTrue(
            np.allclose(
                _read_model_function.func(self._test_x, self._test_a, self._test_b),
                self._test_y
            )
        )
        _read_formatter = _read_model_function.formatter
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        _read_arg_formatters = _read_formatter.arg_formatters
        self.assertTrue(_read_formatter.name == 'linear_model')
        self.assertTrue(_read_formatter.latex_name == 'linear model')
        self.assertTrue(_read_arg_formatters[0].name == 'x')
        self.assertTrue(_read_arg_formatters[0].latex_name == r'{r}')
        self.assertTrue(_read_arg_formatters[1].name == 'alpha')
        self.assertTrue(_read_arg_formatters[1].latex_name == r'{\alpha}')
        self.assertTrue(_read_arg_formatters[2].name == 'beta')
        self.assertTrue(_read_arg_formatters[2].latex_name == r'{\beta}')
        self.assertTrue(_read_formatter.expression_format_string == '{0} * {x} + {1}')
        self.assertTrue(_read_formatter.latex_expression_format_string == '{0}{x} + {1}')

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function, ModelFunctionBase))

        self.assertTrue(
            np.allclose(
                self._test_y,
                _read_model_function.func(self._test_x, self._test_a, self._test_b)
            )
        )
        
        _given_formatter = self._model_function.formatter
        _read_formatter = _read_model_function.formatter
        
        self.assertTrue(isinstance(_read_formatter, ModelFunctionFormatter))
        
        _given_arg_formatters = _given_formatter.arg_formatters
        _read_arg_formatters = _read_formatter.arg_formatters
        
        self.assertTrue(_read_formatter.name == _given_formatter.name)
        self.assertTrue(_read_formatter.latex_name == _given_formatter.latex_name)
        self.assertTrue(_read_arg_formatters[0].name == _given_arg_formatters[0].name)
        self.assertTrue(_read_arg_formatters[0].latex_name == _given_arg_formatters[0].latex_name)
        self.assertTrue(_read_arg_formatters[1].name == _given_arg_formatters[1].name)
        self.assertTrue(_read_arg_formatters[1].latex_name == _given_arg_formatters[1].latex_name)
        self.assertTrue(_read_formatter.expression_format_string ==  _given_formatter.expression_format_string)
        self.assertTrue(_read_formatter.latex_expression_format_string ==  _given_formatter.latex_expression_format_string)



import numpy as np

from .._base import ParametricModelBaseMixin
from .container import UnbinnedContainer
from ..util import function_library

__all__ = ['UnbinnedParametricModel']


class UnbinnedParametricModel(ParametricModelBaseMixin, UnbinnedContainer):
    def __init__(self, data, model_density_function=function_library.normal_distribution,
                 model_parameters=[1.0, 1.0]):

        self.support = np.array(data)

        _model = model_density_function(self.support, *model_parameters)
        if np.isscalar(_model):
            _model = np.ones_like(self.support) * _model

        super(UnbinnedParametricModel, self).__init__(
            # this gets passed to ParametricModelBaseMixin.__init__
            model_func=model_density_function,
            model_parameters=model_parameters,
            # this gets passed to UnbinnedContainer.__init__
            data=_model
        )

    # -- private methods

    def _recalculate(self):
        # use parent class setter for 'data'
        UnbinnedContainer.data.fset(self, self.eval_model_function())
        self._pm_calculation_stale = False

    @property
    def support(self):
        return self._support

    @support.setter
    def support(self, model_support):
        self._support = model_support
        self._pm_calculation_stale = True

    @property
    def data(self):
        if self._pm_calculation_stale:
            self._recalculate()
        return super(UnbinnedParametricModel, self).data

    @data.setter
    def data(self):
        raise TypeError("Parametric model data cannot be set!")

    def eval_model_function(self, support=None, model_parameters=None):
        """
        Evaluate the model function.

        :param support: *x* values of the support points (if ``None``, the model *support* values are used)
        :type support: list or ``None``
        :param model_parameters: values of the model parameters (if ``None``, the current values are used)
        :type model_parameters: list or ``None``
        :return: value(s) of the model function for the given parameters
        :rtype: :py:obj:`numpy.ndarray`
        """
        _x = support if support is not None else self.support
        _pars = model_parameters if model_parameters is not None else self.parameters
        _y = self._model_function_object(_x, *_pars)
        if np.isscalar(_y):
            _y = np.ones_like(_x) * _y
        return _y



"""
:synopsis: This submodule contains the abstract base classes for all objects used by the :py:mod:`kafe2.fit` module.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .container import *
from .cost import *
from .ensemble import *
from .fit import *
from .format import *
from .model import *
from .plot import *



import unittest
from kafe2.test.core.minimizers._base import AbstractMinimizerTest

_cannot_import_ROOT = False
try:
    from kafe2.core.minimizers.root_tminuit_minimizer import \
        MinimizerROOTTMinuit, MinimizerROOTTMinuitException
except ImportError:
    _cannot_import_ROOT = True


@unittest.skipIf(_cannot_import_ROOT, "Cannot import ROOT")
class TestMinimizerROOTTMinuit(AbstractMinimizerTest, unittest.TestCase):

    def _get_minimizer(self, parameter_names, parameter_values, parameter_errors,
                       function_to_minimize):
        return MinimizerROOTTMinuit(
            parameter_names=parameter_names, parameter_values=parameter_values,
            parameter_errors=parameter_errors, function_to_minimize=function_to_minimize
        )

    @property
    def _expected_tolerance(self):
        return 1e-9

    def test_compare_par_values_minimize_fcn3_limit_unlimit_onesided(self):
        with self.assertRaises(MinimizerROOTTMinuitException):
            self.m3.limit('x', (None, 0.7))

    def test_compare_par_values_minimize_fcn3_limit_unlimit_onesided_2(self):
        with self.assertRaises(MinimizerROOTTMinuitException):
            self.m3.limit('x', (3.5, None))



import unittest
import numpy as np

from six import StringIO

from kafe2.fit import IndexedContainer, XYContainer, HistContainer
from kafe2.fit.representation import DataContainerYamlWriter, DataContainerYamlReader
from kafe2.fit.io.handle import IOStreamHandle


# TODO: check that exceptions are raised when appropriate...


TEST_DATASET_INDEXED = """
type: "indexed"
data: [80.429, 80.339]
errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.1
  name: ErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.2
  name: ErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !symmetric_matrix |
    0.3
    0.1   0.3
  name: ErrorThree
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.2]]
  name: ErrorFour
  matrix_type: covariance
  relative: false
  type: matrix
"""

TEST_DATASET_INDEXED_MISSING_KEYWORD = """
type: "indexed"
errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.1
  name: ErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.2
  name: ErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !symmetric_matrix |
    0.3
    0.1   0.3
  name: ErrorThree
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.2]]
  name: ErrorFour
  matrix_type: covariance
  relative: false
  type: matrix
"""

TEST_DATASET_INDEXED_EXTRA_KEYWORD = TEST_DATASET_INDEXED + """
extra_keyword: 3.14
"""

class TestIndexedContainerYamlRepresentation(unittest.TestCase):

    def setUp(self):
        _data = [4, 5, 6, 2, 4, 5., 3.01e-5, 4.4e+45]
        _ndat = len(_data)
        self._container = IndexedContainer(data=_data)

        self._container.add_error(err_val=0.1, name="SUA", correlation=0.0, relative=False)
        self._container.add_error(err_val=0.1, name="SUR", correlation=0.0, relative=True)
        self._container.add_error(err_val=0.1, name="SCA", correlation=1.0, relative=False)
        self._container.add_matrix_error(err_matrix=np.eye(_ndat) * 0.1, name="MCov", relative=False, matrix_type='covariance')
        self._container.add_matrix_error(err_matrix=np.eye(_ndat), name="MCor", relative=False, matrix_type='correlation', err_val=0.1)

        # # xy
        # self._container.add_simple_error(axis='x', err_val=0.1, correlation=0.0, relative=False)
        # self._container.add_simple_error(axis='y', err_val=0.2, correlation=0.0, relative=False)
        # self._container.add_matrix_error(axis='y', err_matrix=np.eye(8) * 0.1, relative=False, matrix_type='covariance')

        # # hist
        # self._container.add_simple_error(err_val=2, correlation=0.5, relative=False)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_DATASET_INDEXED))

        self._roundtrip_streamreader = DataContainerYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = DataContainerYamlWriter(self._container, self._roundtrip_stringstream)
        self._testfile_streamreader = DataContainerYamlReader(self._testfile_stringstream)

        self._ref_testfile_data = [80.429, 80.339]
        self._ref_testfile_err = [0.89442719,  0.89442719]
        self._ref_testfile_cov_mat = np.array([[0.8, 0.3], [0.3, 0.8]])
        self._ref_testfile_error_names = {'ErrorOne', 'ErrorTwo', 'ErrorThree', 'ErrorFour'}

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_DATASET_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_DATASET_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = DataContainerYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = DataContainerYamlReader(self._testfile_stringstream_extra_keyword)


    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()


    def test_read_from_testfile_stream(self):
        _read_container = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_container, IndexedContainer))
        self.assertTrue(
            np.allclose(
                _read_container.data,
                self._ref_testfile_data
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.err,
                self._ref_testfile_err
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.cov_mat,
                self._ref_testfile_cov_mat,
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(_read_container._error_dicts.keys()),
            self._ref_testfile_error_names
        )


    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_container = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_container, IndexedContainer))

        # compare data members
        self.assertTrue(
            np.allclose(
                self._container.data,
                _read_container.data
            )
        )

        # compare (total) errors and cov mats
        # TODO: compare individual error components -> nontrivial because error ids change...
        self.assertTrue(
            np.allclose(
                self._container.err,
                _read_container.err
            )
        )

        self.assertTrue(
            np.allclose(
                self._container.cov_mat,
                _read_container.cov_mat
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(self._container._error_dicts.keys()),
            set(_read_container._error_dicts.keys())
        )


TEST_DATASET_XY = """
type: "xy"
x_data: [5, 17]
y_data: [80.429, 80.339]
y_errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.1
  name: YErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.2
  name: YErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
x_errors:
- matrix: !symmetric_matrix |
    0.3
    0.1   0.3
  name: XErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.2]]
  name: XErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
"""

TEST_DATASET_XY_MISSING_KEYWORD = """
type: "xy"
y_data: [80.429, 80.339]
y_errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.1
  name: YErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.2
  name: YErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
x_errors:
- matrix: !symmetric_matrix |
    0.3
    0.1   0.3
  name: XErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.2]]
  name: XErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
"""

TEST_DATASET_XY_EXTRA_KEYWORD = TEST_DATASET_XY + """
extra_keyword: 3.14
"""


class TestXYContainerYamlRepresentation(unittest.TestCase):

    def setUp(self):
        _data = [[0.0, .1, .3], [10, 24, 44]]
        _ndat = len(_data[0])
        self._container = XYContainer(x_data=_data[0], y_data=_data[1])

        self._container.add_error(axis='y', name="ySUA", err_val=0.1, correlation=0.0, relative=False)
        self._container.add_error(axis='y', name="ySUR", err_val=0.1, correlation=0.0, relative=True)
        self._container.add_error(axis='y', name="ySCA", err_val=0.1, correlation=1.0, relative=False)
        self._container.add_matrix_error(axis='y', name="yMCov", err_matrix=np.eye(_ndat) * 0.1, relative=False, matrix_type='covariance')
        self._container.add_matrix_error(axis='y', name="yMCor", err_matrix=np.eye(_ndat), relative=False, matrix_type='correlation', err_val=0.1)

        self._container.add_error(axis='x', name="xSUA", err_val=0.1, correlation=0.0, relative=False)
        self._container.add_error(axis='x', name="xSUR", err_val=0.1, correlation=0.0, relative=True)
        self._container.add_matrix_error(axis='x', name="xMCov", err_matrix=np.eye(_ndat) * 0.1, relative=False, matrix_type='covariance')

        # self._container.add_simple_error(axis='x', err_val=0.1, correlation=0.0, relative=False)
        # self._container.add_simple_error(axis='y', err_val=0.2, correlation=0.0, relative=False)
        # self._container.add_matrix_error(axis='y', err_matrix=np.eye(8) * 0.1, relative=False, matrix_type='covariance')

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_DATASET_XY))

        self._roundtrip_streamreader = DataContainerYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = DataContainerYamlWriter(self._container, self._roundtrip_stringstream)
        self._testfile_streamreader = DataContainerYamlReader(self._testfile_stringstream)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_DATASET_XY_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_DATASET_XY_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = DataContainerYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = DataContainerYamlReader(self._testfile_stringstream_extra_keyword)

        self._ref_testfile_x_data = [5, 17]
        self._ref_testfile_y_data = [80.429, 80.339]
        self._ref_testfile_x_err = [0.70710678,  0.70710678]
        self._ref_testfile_y_err = [0.54772256,  0.54772256]
        self._ref_testfile_y_cov_mat = np.array([[0.3, 0.1], [0.1, 0.3]])
        self._ref_testfile_x_cov_mat = np.array([[0.5, 0.2], [0.2, 0.5]])
        self._ref_testfile_error_names = {'XErrorOne', 'XErrorTwo', 'YErrorOne', 'YErrorTwo'}


    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()


    def test_read_from_testfile_stream(self):
        _read_container = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_container, XYContainer))
        self.assertTrue(
            np.allclose(
                _read_container.x,
                self._ref_testfile_x_data
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.y,
                self._ref_testfile_y_data
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.x_err,
                self._ref_testfile_x_err
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.y_err,
                self._ref_testfile_y_err
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.x_cov_mat,
                self._ref_testfile_x_cov_mat,
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.y_cov_mat,
                self._ref_testfile_y_cov_mat,
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(_read_container._error_dicts.keys()),
            self._ref_testfile_error_names
        )


    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_container = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_container, XYContainer))

        # compare data members
        self.assertTrue(
            np.allclose(
                self._container.x,
                _read_container.x
            )
        )
        self.assertTrue(
            np.allclose(
                self._container.y,
                _read_container.y
            )
        )

        # compare (total) errors and cov mats
        # TODO: compare individual error components -> nontrivial because error ids change...
        self.assertTrue(
            np.allclose(
                self._container.x_err,
                _read_container.x_err
            )
        )
        self.assertTrue(
            np.allclose(
                self._container.y_err,
                _read_container.y_err
            )
        )

        self.assertTrue(
            np.allclose(
                self._container.x_cov_mat,
                _read_container.x_cov_mat
            )
        )
        self.assertTrue(
            np.allclose(
                self._container.y_cov_mat,
                _read_container.y_cov_mat
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(self._container._error_dicts.keys()),
            set(_read_container._error_dicts.keys())
        )


TEST_DATASET_HIST = """
bin_edges:
- 1.0
- 2.0
- 3.0
raw_data:
- 4
- 5
- 3
- 3
- 2
- 1
errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.2
  name: ErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.3
  name: ErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !symmetric_matrix |
    0.3
    0.1   0.4
  name: ErrorThree
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.1]]
  name: ErrorFour
  matrix_type: covariance
  relative: false
  type: matrix
type: histogram
"""

TEST_DATASET_HIST_EXTRA_KEYWORD = TEST_DATASET_HIST + """
extra_keyword: 3.14
"""

TEST_DATASET_HIST_MANUAL_HEIGHTS = """
bin_edges:
- 1.0
- 2.0
- 3.0
bin_heights:
- 1
- 1
underflow: 0
overflow: 4
errors:
- matrix:
  - - 0.1
    - 0.0
  - - 0.0
    - 0.2
  name: ErrorOne
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !matrix |
    0.2   0.1
    0.1   0.3
  name: ErrorTwo
  matrix_type: covariance
  relative: false
  type: matrix
- matrix: !symmetric_matrix |
    0.3
    0.1   0.4
  name: ErrorThree
  matrix_type: covariance
  relative: false
  type: matrix
- matrix:
    [[0.2,  0.1],
     [0.1,  0.1]]
  name: ErrorFour
  matrix_type: covariance
  relative: false
  type: matrix
type: histogram
"""


class TestHistContainerYamlRepresentation(unittest.TestCase):

    def setUp(self):
        _data = [1, 1, 4, 1, 1, 4, 2, 5, 5, 5, 2, 4, 2, 2, 18, 2, 8, 8, 9]
        _nbins = 3
        self._container = HistContainer(n_bins=_nbins,
                                        bin_range=(0.0, 10.),
                                        fill_data=_data)

        self._container.add_error(err_val=0.1, name='SC5A_1', correlation=0.5, relative=False)
        self._container.add_error(err_val=[i / 10. for i in range(1, _nbins + 1)], name='SC5A_2', correlation=0.5, relative=False)
        self._container.add_matrix_error(err_matrix=np.eye(_nbins) * 0.1, name='MCov', relative=False, matrix_type='covariance')
        self._container.add_matrix_error(err_matrix=np.eye(_nbins), name='MCor', relative=False, matrix_type='correlation', err_val=0.03)

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_DATASET_HIST))

        self._roundtrip_streamreader = DataContainerYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = DataContainerYamlWriter(self._container, self._roundtrip_stringstream)
        self._testfile_streamreader = DataContainerYamlReader(self._testfile_stringstream)

        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_DATASET_HIST_EXTRA_KEYWORD))
        self._testfile_streamreader_extra_keyword = DataContainerYamlReader(self._testfile_stringstream_extra_keyword)

        self._testfile_stringstream_manual_heights = IOStreamHandle(
            StringIO(TEST_DATASET_HIST_MANUAL_HEIGHTS))
        self._testfile_streamreader_manual_heights = DataContainerYamlReader(
            self._testfile_stringstream_manual_heights)

        self._ref_testfile_n_bins = 2
        self._ref_testfile_bin_range = [1.0, 3.0]
        self._ref_testfile_bin_edges = [1.0, 2.0, 3.0]
        self._ref_testfile_data = [1, 1]
        self._ref_testfile_underflow = 0
        self._ref_testfile_overflow = 4
        self._ref_testfile_err = [0.89442719,  1]
        self._ref_testfile_cov_mat = np.array([[0.8, 0.3], [0.3, 1.0]])
        self._ref_testfile_error_names = {'ErrorOne', 'ErrorTwo', 'ErrorThree', 'ErrorFour'}

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_container = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_container, HistContainer))

        self.assertTrue(_read_container.n_bins == self._ref_testfile_n_bins)
        self.assertTrue(
            np.allclose(
                _read_container.bin_range,
                self._ref_testfile_bin_range
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.bin_edges,
                self._ref_testfile_bin_edges
            )
        )

        self.assertTrue(
            np.allclose(
                _read_container.data,
                self._ref_testfile_data
            )
        )
        self.assertEqual(_read_container.underflow, self._ref_testfile_underflow)
        self.assertEqual(_read_container.overflow, self._ref_testfile_overflow)
        self.assertTrue(
            np.allclose(
                _read_container.err,
                self._ref_testfile_err
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.cov_mat,
                self._ref_testfile_cov_mat,
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(_read_container._error_dicts.keys()),
            self._ref_testfile_error_names
        )

    def test_read_from_testfile_stream_manual_heights(self):
        _read_container = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_container, HistContainer))

        self.assertTrue(_read_container.n_bins == self._ref_testfile_n_bins)
        self.assertTrue(
            np.allclose(
                _read_container.bin_range,
                self._ref_testfile_bin_range
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.bin_edges,
                self._ref_testfile_bin_edges
            )
        )

        self.assertTrue(
            np.allclose(
                _read_container.data,
                self._ref_testfile_data
            )
        )
        self.assertEqual(_read_container.underflow, self._ref_testfile_underflow)
        self.assertEqual(_read_container.overflow, self._ref_testfile_overflow)
        self.assertTrue(
            np.allclose(
                _read_container.err,
                self._ref_testfile_err
            )
        )
        self.assertTrue(
            np.allclose(
                _read_container.cov_mat,
                self._ref_testfile_cov_mat,
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(_read_container._error_dicts.keys()),
            self._ref_testfile_error_names
        )

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_container = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_container, HistContainer))

        self.assertTrue(self._container.n_bins == _read_container.n_bins)
        self.assertTrue(
            np.allclose(
                self._container.bin_range,
                _read_container.bin_range
            )
        )
        self.assertTrue(
            np.allclose(
                self._container.bin_edges,
                _read_container.bin_edges
            )
        )

        # compare data members
        self.assertTrue(
            np.allclose(
                self._container.data,
                _read_container.data
            )
        )

        # compare (total) errors and cov mats
        # TODO: compare individual error components -> nontrivial because error ids change...
        self.assertTrue(
            np.allclose(
                self._container.err,
                _read_container.err
            )
        )

        self.assertTrue(
            np.allclose(
                self._container.cov_mat,
                _read_container.cov_mat
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(self._container._error_dicts.keys()),
            set(_read_container._error_dicts.keys())
        )

    def test_round_trip_with_stringstream_manual_heights(self):
        self._container.set_bins(
            self._container.data, self._container.underflow, self._container.overflow)
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_container = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_container, HistContainer))

        self.assertTrue(self._container.n_bins == _read_container.n_bins)
        self.assertTrue(
            np.allclose(
                self._container.bin_range,
                _read_container.bin_range
            )
        )
        self.assertTrue(
            np.allclose(
                self._container.bin_edges,
                _read_container.bin_edges
            )
        )

        # compare data members
        self.assertTrue(
            np.allclose(
                self._container.data,
                _read_container.data
            )
        )

        # compare (total) errors and cov mats
        # TODO: compare individual error components -> nontrivial because error ids change...
        self.assertTrue(
            np.allclose(
                self._container.err,
                _read_container.err
            )
        )

        self.assertTrue(
            np.allclose(
                self._container.cov_mat,
                _read_container.cov_mat
            )
        )

        # check that the error names are the same
        self.assertEqual(
            set(self._container._error_dicts.keys()),
            set(_read_container._error_dicts.keys())
        )



import abc
import six

from .._base import GenericDReprBase
# import formatter classes
from ..._base.format import ModelFunctionFormatter
from ...indexed import IndexedModelFunctionFormatter

__all__ = ["ModelFunctionFormatterDReprBase", "ParameterFormatterDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ModelFunctionFormatterDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'model_function_formatter'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        ModelFunctionFormatter: 'base',
        IndexedModelFunctionFormatter: 'indexed'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'base': ModelFunctionFormatter,
        'histogram': ModelFunctionFormatter,
        'indexed': IndexedModelFunctionFormatter,
        'unbinned': ModelFunctionFormatter,
        'xy': ModelFunctionFormatter,
    }

    def __init__(self, model_function_formatter=None):
        self._kafe_object = model_function_formatter
        super(ModelFunctionFormatterDReprBase, self).__init__()


@six.add_metaclass(abc.ABCMeta)
class ParameterFormatterDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'parameter_formatter'

    def __init__(self, model_parameter_formatter=None):
        self._kafe_object = model_parameter_formatter
        super(ParameterFormatterDReprBase, self).__init__()



from copy import deepcopy

import sys

from .._base import FitBase, DataContainerBase, ModelFunctionBase
from .container import UnbinnedContainer
from .cost import UnbinnedCostFunction_NegLogLikelihood, STRING_TO_COST_FUNCTION
from .model import UnbinnedParametricModel
from .plot import UnbinnedPlotAdapter
from ..util import collect

__all__ = ['UnbinnedFit']


class UnbinnedFit(FitBase):
    CONTAINER_TYPE = UnbinnedContainer
    MODEL_TYPE = UnbinnedParametricModel
    MODEL_FUNCTION_TYPE = ModelFunctionBase
    PLOT_ADAPTER_TYPE = UnbinnedPlotAdapter
    RESERVED_NODE_NAMES = {'data', 'model', 'cost', 'parameter_values', 'parameter_constraints'}

    _STRING_TO_COST_FUNCTION = STRING_TO_COST_FUNCTION

    def __init__(self,
                 data,
                 model_function='normal_distribution',
                 cost_function=UnbinnedCostFunction_NegLogLikelihood(),
                 minimizer=None,
                 minimizer_kwargs=None):
        """
        Construct a fit to a model of *unbinned* data.

        :param data: the data points
        :param model_function: the model density
        :type model_function: :py:class:`~kafe2.fit._base.ModelFunctionBase` or unwrapped native Python function
        :param cost_function: the cost function
        :param minimizer: the minimizer to use for fitting.
        :type minimizer: None, "iminuit", "tminuit", or "scipy".
        :param minimizer_kwargs: dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        super(UnbinnedFit, self).__init__(
            data=data, model_function=model_function, cost_function=cost_function,
            minimizer=minimizer, minimizer_kwargs=minimizer_kwargs)

    # private methods

    def _init_nexus(self):
        super(UnbinnedFit, self)._init_nexus()

        # add 'x' as an alias of 'data'
        self._nexus.add_alias('x', alias_for='data')

        self._nexus.add_dependency(
            'model',
            depends_on=(
                'parameter_values'
            )
        )

    # -- private methods

    def _set_new_data(self, new_data):
        if isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '%s' (expected '%s')"
                                       % (type(new_data), self.CONTAINER_TYPE))
        else:
            self._data_container = UnbinnedContainer(new_data, dtype=float)
        self._data_container._on_error_change_callback = self._on_error_change

        self._nexus.get('x').mark_for_update()
        # TODO: make 'Alias' nodes pass on 'mark_for_update'
        self._nexus.get('data').mark_for_update()

    def _set_new_parametric_model(self):
        self._param_model = UnbinnedParametricModel(
            data=self.data,
            model_density_function=self._model_function,
            model_parameters=self.parameter_values
        )

    @property
    def data_range(self):
        """The minimum and maximum value of the data"""
        return self._data_container.data_range

    @property
    def model(self):
        """array of model predictions for the data points"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.data

    @property
    def goodness_of_fit(self):
        return None

    def eval_model_function(self, x=None, model_parameters=None):
        """
        Evaluate the model function.

        :param x: values of *x* at which to evaluate the model function (if ``None``, the data *x* values are used)
        :type x: iterable of float
        :param model_parameters: the model parameter values (if ``None``, the current values are used)
        :type model_parameters: iterable of float
        :return: model function values
        :rtype: :py:class:`numpy.ndarray`
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        self._param_model.support = self.data
        return self._param_model.eval_model_function(support=x, model_parameters=model_parameters)

    def report(self, output_stream=sys.stdout, asymmetric_parameter_errors=False):
        super(UnbinnedFit, self).report(output_stream=output_stream,
                                        asymmetric_parameter_errors=asymmetric_parameter_errors)



from .._base import CostFunction_Chi2, CostFunction_NegLogLikelihood, \
    CostFunction_GaussApproximation

__all__ = [
    "XYCostFunction_Chi2",
    "XYCostFunction_NegLogLikelihood",
    "XYCostFunction_GaussApproximation"
]


class XYCostFunction_Chi2(CostFunction_Chi2):
    def __init__(
            self, errors_to_use='covariance', fallback_on_singular=True, axes_to_use='xy',
            add_constraint_cost=True, add_determinant_cost=True):
        """Built-in least-squares cost function for *xy* data.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`. This is either
            `'covariance'``, ``'pointwise'`` or :py:obj:`None`.
        :type errors_to_use: str or None
        :param axes_to_use: The errors for the given axes are taken into account when calculating
            :math:`\\chi^2`. Either ``'y'`` or ``'xy'``
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == 'y':
            self._COV_MAT_CHOLESKY_NAME = "y_total_cov_mat_cholesky"
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == 'xy':
            self._COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError(
                "Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super(XYCostFunction_Chi2, self).__init__(
            errors_to_use=errors_to_use, fallback_on_singular=fallback_on_singular,
            add_constraint_cost=add_constraint_cost, add_determinant_cost=add_determinant_cost)


class XYCostFunction_NegLogLikelihood(CostFunction_NegLogLikelihood):
    def __init__(self, data_point_distribution="poisson", ratio=False, axes_to_use="xy"):
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == 'y':
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == 'xy':
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError(
                "Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super(XYCostFunction_NegLogLikelihood, self).__init__(
            data_point_distribution=data_point_distribution, ratio=ratio)


class XYCostFunction_GaussApproximation(CostFunction_GaussApproximation):
    def __init__(
            self, errors_to_use='covariance', axes_to_use='xy',
            add_constraint_cost=True, add_determinant_cost=True):
        """
        Built-in Gaussian approximation of the Poisson negative log-likelihood cost function for
        *xy* data.

        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`. This is either
            `'covariance'``, ``'pointwise'``.
        :type errors_to_use: str
        :param axes_to_use: The errors for the given axes are taken into account when calculating
            :math:`\\chi^2`. Either ``'y'`` or ``'xy'``
        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
            constraints.
        :param bool add_determinant_cost: If :py:obj:`True`, automatically increase the cost
            function value by the logarithm of the determinant of the covariance matrix to reduce
            bias.
        """
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        if axes_to_use.lower() == 'y':
            self._COV_MAT_CHOLESKY_NAME = "y_total_cov_mat_cholesky"
            self._ERROR_NAME = "y_total_error"
        elif axes_to_use.lower() == 'xy':
            self._COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
            self._ERROR_NAME = "total_error"
        else:
            raise ValueError(
                "Unknown value '%s' for 'axes_to_use': must be one of ('xy', 'y')")
        super().__init__(
            errors_to_use=errors_to_use, add_constraint_cost=add_constraint_cost,
            add_determinant_cost=add_determinant_cost
        )


STRING_TO_COST_FUNCTION = {
    'chi2': (XYCostFunction_Chi2, {}),
    'chi_2': (XYCostFunction_Chi2, {}),
    'chisquared': (XYCostFunction_Chi2, {}),
    'chi_squared': (XYCostFunction_Chi2, {}),
    'chi2_no_errors': (XYCostFunction_Chi2, {"errors_to_use": None, "add_determinant_cost": False}),
    'chi2_pointwise': (XYCostFunction_Chi2, {"errors_to_use": "pointwise"}),
    'chi2_pointwise_errors': (XYCostFunction_Chi2, {"errors_to_use": "pointwise"}),
    'chi2_covariance': (XYCostFunction_Chi2, {"errors_to_use": "covariance"}),
    'nll': (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    'poisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nll-poisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nll_poisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nllpoisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": False}),
    'nll-gaussian': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'nll_gaussian': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'nllgaussiann': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": False}),
    'negloglikelihood': (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    'neg_log_likelihood': (XYCostFunction_NegLogLikelihood, {"ratio": False}),
    'nllr': (CostFunction_NegLogLikelihood, {"ratio": True}),
    'nllr-poisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllr_poisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllrpoisson': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "poisson", "ratio": True}),
    'nllr-gaussian': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'nllr_gaussian': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'nllrgaussian': (XYCostFunction_NegLogLikelihood,
                    {"data_point_distribution": "gaussian", "ratio": True}),
    'negloglikelihoodratio': (XYCostFunction_NegLogLikelihood, {"ratio": True}),
    'neg_log_likelihood_ratio': (XYCostFunction_NegLogLikelihood, {"ratio": True}),
    'gauss-approximation': (XYCostFunction_GaussApproximation, {}),
    'gauss_approximation': (XYCostFunction_GaussApproximation, {}),
    'gauss_approximation_covariance': (
        XYCostFunction_GaussApproximation, {"errors_to_use": "covariance"}),
    'gauss_approximation_pointwise': (
        XYCostFunction_GaussApproximation, {"errors_to_use": "pointwise"}),
    'gauss_approximation_pointwise_errors': (
        XYCostFunction_GaussApproximation, {"errors_to_use": "pointwise"}),
}



import unittest
from six import StringIO

from kafe2.fit._base import ParameterFormatter, ModelFunctionFormatter
from kafe2.fit.indexed import IndexedModelFunctionFormatter
from kafe2.fit.representation import ModelFunctionFormatterYamlWriter,\
    ModelFunctionFormatterYamlReader, ParameterFormatterYamlWriter, ParameterFormatterYamlReader
from kafe2.fit.io.handle import IOStreamHandle


TEST_MODEL_FUNCTION_FORMATTER = """
name: quadratic_model
latex_name: quadratic model
signature: [x, a, b, c]
arg_formatters:
  x: X
  a: A
  b: B
  c: C
expression_string: '{0} * {x} ** 2 + {1} * {x} + {2}'
latex_expression_string: '{0}{x}^2 + {1}{x} + {2}' 
"""

TEST_MODEL_FUNCTION_FORMATTER_MISSING_KEYWORD = """
latex_name: quadratic model
signature: [x, a, b, c]
arg_formatters:
  x: X
  a: A
  b: B
  c: C
expression_string: '{0} * {x} ** 2 + {1} * {x} + {2}'
latex_expression_string: '{0}{x}^2 + {1}{x} + {2}' 
"""

TEST_MODEL_FUNCTION_FORMATTER_EXTRA_KEYWORD = TEST_MODEL_FUNCTION_FORMATTER + """\nextra_keyword: 3.14\n"""


class TestModelFunctionFormatterYamlRepresenter(unittest.TestCase):

    def setUp(self):
        self._model_function_formatter = ModelFunctionFormatter(
            name='quadratic_model',
            latex_name='quadratic model',
            arg_formatters=[
                ParameterFormatter(arg_name='x', latex_name='X'),
                ParameterFormatter(arg_name='a', value=1.1, error=0.1, latex_name='A'),
                ParameterFormatter(arg_name='b', value=2.2, error=0.1, latex_name='B'),
                ParameterFormatter(arg_name='c', value=3.3, error=0.1, latex_name='C')
            ],
            expression_string='{0} * {x} ** 2 + {1} * {x} + {2}',
            latex_expression_string='{0}{x}^2 + {1}{x} + {2}' 
        )
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER))

        self._roundtrip_streamreader = ModelFunctionFormatterYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionFormatterYamlWriter(self._model_function_formatter,
                                                                        self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionFormatterYamlReader(self._testfile_stringstream)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionFormatterYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionFormatterYamlReader(self._testfile_stringstream_extra_keyword)
    
    def _assert_model_function_formatters_equal(self, formatter1, formatter2):
        self.assertTrue(formatter1.name == formatter2.name)
        self.assertTrue(formatter1.latex_name == formatter2.latex_name)
        for _arg_formatter_1, _arg_formatter_2 in zip(
                formatter1._arg_formatters, formatter2._arg_formatters):
            self.assertTrue(_arg_formatter_1.name == _arg_formatter_2.name)
            self.assertTrue(_arg_formatter_1.latex_name == _arg_formatter_2.latex_name)
        self.assertTrue(formatter1.expression_format_string == formatter2.expression_format_string)
        self.assertTrue(formatter1.latex_expression_format_string == formatter2.latex_expression_format_string)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function_formatter = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function_formatter, ModelFunctionFormatter))
        self._assert_model_function_formatters_equal(_read_model_function_formatter, self._model_function_formatter)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function_formatter = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function_formatter, ModelFunctionFormatter))
        self._assert_model_function_formatters_equal(_read_model_function_formatter, self._model_function_formatter)


TEST_MODEL_FUNCTION_FORMATTER_INDEXED = """
type: indexed
name: quadratic_model
latex_name: quadratic model
index_name: x
latex_index_name: X
signature: [a, b, c]
arg_formatters:
  a: A
  b: B
  c: C
expression_string: '{0} * {x} ** 2 + {1} * {x} + {2}'
latex_expression_string: '{0}{x}^2 + {1}{x} + {2}' 
"""

TEST_MODEL_FUNCTION_FORMATTER_INDEXED_MISSING_KEYWORD = """
type: indexed
latex_name: quadratic model
index_name: x
latex_index_name: X
signature: [a, b, c]
arg_formatters:
  a: A
  b: B
  c: C
expression_string: '{0} * {x} ** 2 + {1} * {x} + {2}'
latex_expression_string: '{0}{x}^2 + {1}{x} + {2}' 
"""

TEST_MODEL_FUNCTION_FORMATTER_INDEXED_EXTRA_KEYWORD = TEST_MODEL_FUNCTION_FORMATTER_INDEXED + """
extra_keyword: 3.14
"""


class TestIndexedModelFunctionFormatterYamlRepresenter(unittest.TestCase):

    def setUp(self):
        self._model_function_formatter = IndexedModelFunctionFormatter(
            name='quadratic_model',
            latex_name='quadratic model',
            index_name='x',
            latex_index_name='X',
            arg_formatters=[
                ParameterFormatter(arg_name='a', value=1.1, error=0.1, latex_name='A'),
                ParameterFormatter(arg_name='b', value=2.2, error=0.1, latex_name='B'),
                ParameterFormatter(arg_name='c', value=3.3, error=0.1, latex_name='C')
            ],
            expression_string='{0} * {x} ** 2 + {1} * {x} + {2}',
            latex_expression_string='{0}{x}^2 + {1}{x} + {2}' 
        )
        
        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER_INDEXED))

        self._roundtrip_streamreader = ModelFunctionFormatterYamlReader(self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ModelFunctionFormatterYamlWriter(self._model_function_formatter,
                                                                        self._roundtrip_stringstream)
        self._testfile_streamreader = ModelFunctionFormatterYamlReader(self._testfile_stringstream)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER_INDEXED_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(StringIO(TEST_MODEL_FUNCTION_FORMATTER_INDEXED_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ModelFunctionFormatterYamlReader(self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ModelFunctionFormatterYamlReader(self._testfile_stringstream_extra_keyword)
    
    def _assert_model_function_formatters_equal(self, formatter1, formatter2):
        self.assertTrue(formatter1.name == formatter2.name)
        self.assertTrue(formatter1.latex_name == formatter2.latex_name)
        self.assertTrue(formatter1.index_name == formatter2.index_name)
        self.assertTrue(formatter1.latex_index_name == formatter2.latex_index_name)
        for _arg_formatter_1, _arg_formatter_2 in zip(
                formatter1._arg_formatters, formatter2._arg_formatters):
            self.assertTrue(_arg_formatter_1.name == _arg_formatter_2.name)
            self.assertTrue(_arg_formatter_1.latex_name == _arg_formatter_2.latex_name)
        self.assertTrue(formatter1.expression_format_string == formatter2.expression_format_string)
        self.assertTrue(formatter1.latex_expression_format_string == formatter2.latex_expression_format_string)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_function_formatter = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_function_formatter, IndexedModelFunctionFormatter))
        self._assert_model_function_formatters_equal(_read_model_function_formatter, self._model_function_formatter)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_model_function_formatter = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_model_function_formatter, IndexedModelFunctionFormatter))
        self._assert_model_function_formatters_equal(_read_model_function_formatter, self._model_function_formatter)


TEST_MODEL_PARAMETER_FORMATTER = """
id: phi
name: ph
latex_name: \phi
"""

TEST_MODEL_PARAMETER_FORMATTER_MISSING_KEYWORD = """
latex_name: \phi
"""

TEST_MODEL_PARAMETER_FORMATTER_EXTRA_KEYWORD = """
name: phi
latex_name: \phi
extra_keyword: 3.14
"""


class TestModelParameterFormatterYamlRepresenter(unittest.TestCase):

    def setUp(self):
        self._model_parameter_formatter = ParameterFormatter('phi',
            name='ph',
            value=1.571,
            error=0.1,
            latex_name=r"\phi"
        )

        self._roundtrip_stringstream = IOStreamHandle(StringIO())
        self._testfile_stringstream = IOStreamHandle(StringIO(TEST_MODEL_PARAMETER_FORMATTER))

        self._roundtrip_streamreader = ParameterFormatterYamlReader(
            self._roundtrip_stringstream)
        self._roundtrip_streamwriter = ParameterFormatterYamlWriter(
            self._model_parameter_formatter,
            self._roundtrip_stringstream)
        self._testfile_streamreader = ParameterFormatterYamlReader(self._testfile_stringstream)

        self._testfile_stringstream_missing_keyword = IOStreamHandle(
            StringIO(TEST_MODEL_PARAMETER_FORMATTER_MISSING_KEYWORD))
        self._testfile_stringstream_extra_keyword = IOStreamHandle(
            StringIO(TEST_MODEL_PARAMETER_FORMATTER_EXTRA_KEYWORD))
        self._testfile_streamreader_missing_keyword = ParameterFormatterYamlReader(
            self._testfile_stringstream_missing_keyword)
        self._testfile_streamreader_extra_keyword = ParameterFormatterYamlReader(
            self._testfile_stringstream_extra_keyword)

    def test_write_to_roundtrip_stringstream(self):
        self._roundtrip_streamwriter.write()

    def test_read_from_testfile_stream(self):
        _read_model_parameter_formatter = self._testfile_streamreader.read()
        self.assertTrue(isinstance(_read_model_parameter_formatter, ParameterFormatter))

        self.assertTrue(
            _read_model_parameter_formatter.name == self._model_parameter_formatter.name)
        self.assertTrue(_read_model_parameter_formatter.value is None)
        self.assertTrue(_read_model_parameter_formatter.error is None)
        self.assertTrue(
            _read_model_parameter_formatter.latex_name == self._model_parameter_formatter.latex_name)

    def test_read_from_testfile_stream_missing_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_missing_keyword.read()

    def test_read_from_testfile_stream_extra_keyword(self):
        with self.assertRaises(ValueError):
            self._testfile_streamreader_extra_keyword.read()

    def test_round_trip_with_stringstream(self):
        self._roundtrip_streamwriter.write()
        self._roundtrip_stringstream.seek(0)  # return to beginning
        _read_parameter_formatter = self._roundtrip_streamreader.read()
        self.assertTrue(isinstance(_read_parameter_formatter, ParameterFormatter))

        self.assertTrue(_read_parameter_formatter.name == self._model_parameter_formatter.name)
        self.assertTrue(_read_parameter_formatter.value is None)
        self.assertTrue(_read_parameter_formatter.error is None)
        self.assertTrue(
            _read_parameter_formatter.latex_name == self._model_parameter_formatter.latex_name)



import matplotlib as mpl
import numpy as np
import six

from .._base import PlotAdapterBase
from .._base.plot import kc_plot_style

from ..xy.plot import XYPlotAdapter

__all__ = ["HistPlotAdapter"]


class HistPlotAdapter(PlotAdapterBase):

    PLOT_STYLE_CONFIG_DATA_TYPE = 'histogram'

    PLOT_SUBPLOT_TYPES = dict(
        PlotAdapterBase.PLOT_SUBPLOT_TYPES,
        model_density=dict(
            plot_adapter_method='plot_model_density',
            target_axes='main',
        ),
    )

    AVAILABLE_X_SCALES = ('linear', 'log')

    def __init__(self, hist_fit_object, from_container=False):
        """
        Construct an :py:obj:`HistPlotContainer` for a :py:obj:`~kafe2.fit.histogram.HistFit` object:

        :param fit_object: an :py:obj:`~kafe2.fit.histogram.HistFit` object
        :param n_plot_points_model_density: number of plot points to use for plotting the model density
        :param from_container: Whether hist_fit_object was created ad-hoc from just a data
            container.
        :type from_container: bool
        """
        super(HistPlotAdapter, self).__init__(
            fit_object=hist_fit_object, from_container=from_container)
        self.n_plot_points = 100 if len(self.data_x) < 100 else len(self.data_x)
        self.x_range = self._fit.data_container.bin_range

    # -- private methods

    def _set_plot_labels(self):
        super(HistPlotAdapter, self)._set_plot_labels()
        # set model density label according to model label
        _model_label = self._fit.model_label
        if _model_label is None:
            _model_label = dict(kc_plot_style(self.PLOT_STYLE_CONFIG_DATA_TYPE, 'model', 'plot_kwargs'))['label']
        _density_label = dict(kc_plot_style(self.PLOT_STYLE_CONFIG_DATA_TYPE, 'model_density', 'plot_kwargs'))['label']
        _density_label = _density_label % dict(model_label=_model_label)
        self.update_plot_kwargs('model_density', dict(label=_density_label))

    # -- public methods

    @property
    def data_x(self):
        """data x values"""
        return self._fit.data_container.bin_centers

    @property
    def data_y(self):
        """data y values"""
        return self._fit.data.astype(float)

    @property
    def data_xerr(self):
        """x error bars for data (actually used to represent the bins)"""
        return self._fit.data_container.bin_widths * 0.5

    @property
    def data_yerr(self):
        """y error bars for data: total uncertainty"""
        return self._fit.total_error

    @property
    def model_x(self):
        """model prediction x values"""
        return self.data_x

    @property
    def model_y(self):
        """model prediction y values"""
        return self._fit.model

    @property
    def model_xerr(self):
        """x error bars for model (actually used to represent the bins)"""
        return self._fit._param_model.bin_widths * 0.5

    @property
    def model_yerr(self):
        """y error bars for model: ``None`` for :py:obj:`HistPlotContainer`"""
        return None #self._fit.model_error

    @property
    def model_density_x(self):
        """x support points for model density plot"""
        _xmin, _xmax = self.x_range
        if self.x_scale == 'linear':
            return np.linspace(_xmin, _xmax, self.n_plot_points)
        if self.x_scale == 'log':
            try:
                return np.geomspace(_xmin, _xmax, self.n_plot_points)
            except ValueError as _e:
                raise ValueError("Support point calculation failed. The plot range can't include 0"
                                 "when using log scale.") from _e
        raise ValueError("x_range has to be one of {}. Found {} instead.".format(
            self.AVAILABLE_X_SCALES, self.x_scale))

    @property
    def model_density_y(self):
        """value of model density at the support points"""
        _hist_cont = self._fit.data_container
        _mean_bin_size = float(_hist_cont.high - _hist_cont.low)/_hist_cont.size
        _factor = _hist_cont.n_entries * _mean_bin_size
        if self._fit.density:
            return _factor * self._fit.eval_model_function_density(x=self.model_density_x)
        else:
            return _mean_bin_size * self._fit.eval_model_function_density(x=self.model_density_x)

    # public methods

    def plot_data(self, target_axes, **kwargs):
        """
        Plot the measurement data to a specified ``matplotlib`` ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the ``matplotlib`` method ``errorbar``
        :return: plot handle(s)
        """
        _yerr = np.sqrt(
            self.data_yerr ** 2 + self._fit._cost_function.get_uncertainty_gaussian_approximation(self.data_y) ** 2
        )
        return target_axes.errorbar(self.data_x,
                                    self.data_y,
                                    xerr=self.data_xerr,
                                    yerr=_yerr,
                                    **kwargs)

    def plot_model(self, target_axes, **kwargs):
        """
        Plot the model predictions to a specified matplotlib ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the ``matplotlib`` method ``bar``
        :return: plot handle(s)
        """
        #_pad = kwargs.pop('bar_width_pad')
        _sf = kwargs.pop('bar_width_scale_factor')

        # default call signature (matplotlib >= 2)
        _call_dict = dict(
            x=self.model_x,
            align='center',
            height=self.model_y,
            width=self.model_xerr*2.0 * _sf,
            bottom=None,
            **kwargs
        )

        # adjust call signature for matplotlib < 2
        if mpl.__version__.startswith('1'):
            _call_dict['left'] = _call_dict.pop('x')

        return target_axes.bar(**_call_dict)

    def plot_model_density(self, target_axes, **kwargs):
        """
        Plot the model density to a specified ``matplotlib`` ``Axes`` object.

        :param target_axes: ``matplotlib`` ``Axes`` object
        :param kwargs: keyword arguments accepted by the ``matplotlib`` method ``plot``
        :return: plot handle(s)
        """
        # TODO: how to handle/display "error" on the model density?
        return target_axes.plot(self.model_density_x,
                                self.model_density_y,
                                **kwargs)



from collections import OrderedDict
import numpy as np

from ...config import kc
from ..minimizers import get_minimizer
from.nexus import Nexus


class NexusFitter(object):

    def __init__(self, nexus, parameters_to_fit, parameter_to_minimize, minimizer=None,
                 minimizer_kwargs=None):
        """Handles the minimizer and interfacing of the data to it.

        :param Nexus nexus: A kafe2 nexus object used to manage the caching of intermediate
               results for the calculation of the cost function value.
        :param parameters_to_fit:
        :param str parameter_to_minimize: Name of the parameter to minimize. In most cases this
                                          is the cost function value.
        :param minimizer: Name of the minimizer to use.
        :type minimizer: str or None
        :param minimizer_kwargs: Dictionary containing keyword arguments for the minimizer
                                 initialization.
        :type minimizer_kwargs: dict or None
        """
        self._nx = nexus

        self.parameters_to_fit = parameters_to_fit
        self.parameter_to_minimize = parameter_to_minimize

        _minimizer_class = get_minimizer(minimizer)
        if minimizer_kwargs is None:
            minimizer_kwargs = dict()

        # initialize minimizer
        _par_name_val_map = self.get_fit_parameter_values()
        _par_values = [_par_name_val_map[_pn] for _pn in parameters_to_fit]
        self._minimizer = _minimizer_class(
            parameters_to_fit,
            _par_values,
            [0.1 if _v == 0 else 0.1*_v for _v in np.abs(_par_values)],
            self._fcn_wrapper,
            **minimizer_kwargs
        )

        self._fixed_pars = dict()
        self._limited_pars = dict()

        # flags
        self.__minimizing = False  # minimization ongoing?
        self.__state_is_from_minimizer = False

    # -- private methods

    def _get_pars_from_nexus(self, par_names):
        """return list of nexus nodes for `par_names`"""
        _not_found = []
        _pars = []
        for _pn in par_names:
            _par = self._nx.get(_pn)
            if _par is None:
                _not_found.append(_pn)
            else:
                _pars.append(_par)

        if _not_found:
            raise ValueError(
                "Parameters not registered in Nexus: {}".format(
                    ', '.join(_not_found)
                )
            )
        return _pars

    def _minimize(self, max_calls=None):
        """run minimizer"""

        if max_calls is None:
            max_calls = kc('core', 'fitters', 'nexus_fitter', 'max_calls')

        self.__minimizing = True
        self._minimizer.minimize(max_calls=max_calls)
        self.__minimizing = False

        # evaluate function one more time with the final parameters,
        # in order to ensure the nexus is up to date
        _par_vals = self._minimizer.parameter_values
        self._fcn_wrapper(*_par_vals)

        self.__state_is_from_minimizer = True

    def _fcn_wrapper(self, *fit_par_value_list):
        # set fit parameter values
        assert(len(fit_par_value_list) == len(self._fit_pars))
        for _par, _new_value in zip(self._fit_pars, fit_par_value_list):
            _par.value = _new_value

        # evaluate function and return value
        return self._min_par.value

    # -- public properties

    @property
    def minimizer(self):
        return self._minimizer

    @property
    def parameters_to_fit(self):
        return self._fit_par_names

    @parameters_to_fit.setter
    def parameters_to_fit(self, fit_parameters):
        self._fit_pars = self._get_pars_from_nexus(fit_parameters)
        self._fit_par_names = tuple(fit_parameters)

    @property
    def parameter_to_minimize(self):
        return self._min_par_name

    @parameter_to_minimize.setter
    def parameter_to_minimize(self, parameter_to_minimize):
        self._min_par = \
            self._get_pars_from_nexus([parameter_to_minimize])[0]
        self._min_par_name = parameter_to_minimize

    @property
    def fit_parameter_cov_mat(self):
        return self._minimizer.cov_mat

    @property
    def fit_parameter_cor_mat(self):
        return self._minimizer.cor_mat

    @property
    def fit_parameter_errors(self):
        return self._minimizer.parameter_errors

    @fit_parameter_errors.setter
    def fit_parameter_errors(self, new_errors):
        self._minimizer.parameter_errors = new_errors

    @property
    def asymmetric_fit_parameter_errors(self):
        return self._minimizer.asymmetric_parameter_errors

    @property
    def asymmetric_fit_parameter_errors_if_calculated(self):
        return self._minimizer.asymmetric_parameter_errors_if_calculated

    @property
    def parameter_to_minimize_value(self):
        return self._nx.get(self._min_par_name).value

    @property
    def n_fit_par(self):
        return len(self.parameters_to_fit)

    @property
    def state_is_from_minimizer(self):
        return self.__state_is_from_minimizer

    @property
    def fixed_parameters(self):
        return self._fixed_pars.copy()

    @property
    def limited_parameters(self):
        return self._limited_pars.copy()

    # -- public methods

    def do_fit(self):
        self._minimize()

    def fix_parameter(self, name, value=None):
        if value is not None:
            self.set_fit_parameter_values(**{name: value})

        self._minimizer.fix(name)
        _fixed_par_dict = self.get_fit_parameter_values([name])
        self._fixed_pars.update(_fixed_par_dict)

    def release_parameter(self, name):
        self._minimizer.release(name)
        self._fixed_pars.pop(name, None)

    def limit_parameter(self, name, limits):
        self._minimizer.limit(name, limits)
        self._limited_pars.update({name: limits})

    def unlimit_parameter(self, name):
        self._minimizer.unlimit(name)
        self._limited_pars.pop(name, None)

    def contour(self, parameter_name_1, parameter_name_2, sigma=1.0, **kwargs):
        if not self.__state_is_from_minimizer:
            raise RuntimeError(
                "To calculate a contour the do_fit method has to be called first."
            )

        return self._minimizer.contour(parameter_name_1, parameter_name_2, sigma=sigma, **kwargs)

    def profile(self, parameter_name, bins=20, bound=2, args=None, subtract_min=False):
        if not self.__state_is_from_minimizer:
            raise RuntimeError(
                "To calculate a profile the do_fit method has to be called first."
            )

        return self._minimizer.profile(parameter_name, bins=bins, bound=bound, subtract_min=subtract_min)

    def get_fit_parameter_values(self, parameter_names=None):
        if parameter_names is None:
            parameter_names = self._fit_par_names
        return OrderedDict([
            (_pn, self._nx.get(_pn).value)
            for _pn in parameter_names
        ])

    def set_fit_parameter_values(self, **parameter_value_dict):
        _dict_key_set = set(parameter_value_dict.keys())
        _par_name_set = set(self.parameters_to_fit)

        # test parameter names
        if not _dict_key_set.issubset(_par_name_set):
            _unknown_par_names = _dict_key_set - _par_name_set
            raise ValueError("Cannot set fit parameter values: Unknown fit parameters: %r!"
                                       % (_unknown_par_names,))

        # set values in nexus
        for _par_name, _new_value in parameter_value_dict.items():
            self._nx.get(_par_name).value = _new_value
            self._minimizer.set(_par_name, _new_value)

        # set flags
        self.__state_is_from_minimizer = False

    def set_all_fit_parameter_values(self, fit_par_value_list):
        # test list length
        if not len(fit_par_value_list) == len(self.parameters_to_fit):
            raise ValueError(
                "Cannot set all fit parameter values: "
                "{} fit parameters declared, "
                "but {} provided!".format(
                    self.n_fit_par,
                    len(fit_par_value_list)
                )
            )

        # set values in nexus and minimizer
        for _par_name, _par, _new_value in zip(self._fit_par_names, self._fit_pars, fit_par_value_list):
            _par.value = _new_value
            self._minimizer.set(_par_name, _new_value)

        # set flags
        self.__state_is_from_minimizer = False

    def reset_minimizer(self):
        self._minimizer.reset()



import yaml

from ._base import DReprReaderMixin, DReprWriterMixin


class YamlWriterMixin(DReprWriterMixin):
    DREPR_FLAVOR_NAME = 'yaml'
    DUMPER = yaml.Dumper

    """
    A "mixin" class for creating a *yaml* representation writer.
    Inheriting from this class in addition to a DRepr class for
    a particular object type adds methods for writing a yaml document 
    to an output stream.

    Derived classes should inherit from :py:class:`YamlWriterMixin` and the
    relevant ``DRepr`` class (in that order).
    """

    @staticmethod
    def _to_float(property_dict, property_name):
        property_value = property_dict.get(property_name, None)
        if property_value is not None:
            property_dict[property_name] = float(property_value)
        return property_dict

    def write(self):
        self._yaml_doc = self._make_representation(self._kafe_object)
        with self._ohandle as _h:
            try:
                # try to truncate the file to 0 bytes
                _h.truncate(0)
            except IOError:
                # if truncate not available, ignore
                pass
            _h.write(self._get_preface_comment())
            yaml.dump(self._yaml_doc, _h, default_flow_style=False, sort_keys=False)


class YamlReaderMixin(DReprReaderMixin):
    DREPR_FLAVOR_NAME = 'yaml'
    LOADER = yaml.Loader  # TODO SafeLoader
    """
    A "mixin" class for creating a *yaml* representation writer.
    Inheriting from this class in addition to a DRepr class for
    a particular object type adds methods for writing to
    an output stream.

    Derived classes should inherit from :py:class:`YamlReaderMixin` and the
    relevant ``DRepr`` class (in that order).
    """

    @classmethod
    def _make_object(cls, yaml_doc, default_type='xy', **modify_kwargs):
        # strings may be used as shortcuts for frequently used functionality
        if isinstance(yaml_doc, str):
            yaml_doc = cls._process_string(yaml_doc, default_type)
        _overriden_yaml_doc = cls._check_required_keywords_and_override_subspaces(
            yaml_doc, default_type, modify_kwargs)
        _object, _leftover_yaml_doc = cls._convert_yaml_doc_to_object(_overriden_yaml_doc.copy())
        for _keyword in cls._get_ignored_if_none_keywords():
            if _keyword in _leftover_yaml_doc and _leftover_yaml_doc[_keyword] is None:
                _leftover_yaml_doc.pop(_keyword, None)
        if _leftover_yaml_doc:
            raise ValueError("Received unknown or unsupported keywords for constructing a "
                                      "%s object: %s" % (cls.BASE_OBJECT_TYPE_NAME,
                                                         list(_leftover_yaml_doc.keys())))
        return _object

    @classmethod
    def _check_required_keywords_and_override_subspaces(
            cls, yaml_doc, default_type='xy', modify_kwargs=None):
        if modify_kwargs is None:
            modify_kwargs = {}
        if not cls._type_required():
            _kafe_object_class = None
        else:
            _object_type = yaml_doc.get('type', default_type)
            yaml_doc.update(type=_object_type)
            _kafe_object_class = cls._OBJECT_TYPE_NAME_TO_CLASS.get(_object_type, None)
            if _kafe_object_class is None:
                raise TypeError("%s type unknown or not supported: %s"
                                          % (cls.BASE_OBJECT_TYPE_NAME, _object_type))

        yaml_doc = cls._modify_yaml_doc(yaml_doc.copy(), _kafe_object_class, **modify_kwargs)

        _override_dict = cls._get_subspace_override_dict(_kafe_object_class)
        for _keyword in list(_override_dict.keys()):
            _value = yaml_doc.pop(_keyword, None)
            if _value:
                _target_namespaces = _override_dict[_keyword]
                if not isinstance(_target_namespaces, list):
                    _target_namespaces = [_target_namespaces]
                for _target_namespace in _target_namespaces:
                    if _target_namespace not in yaml_doc:
                        yaml_doc[_target_namespace] = dict()
                    yaml_doc[_target_namespace][_keyword] = _value

        _missing_keywords = [_keyword for _keyword in
                             cls._get_required_keywords(yaml_doc, _kafe_object_class)
                             if _keyword not in yaml_doc]
        if _missing_keywords:
            # TODO rework
            raise ValueError("Missing required information for reading in a %s object: %s"
                                      % (_kafe_object_class, _missing_keywords))

        return yaml_doc

    @classmethod
    def _type_required(cls):
        return True

    @classmethod
    def _process_string(cls, string_representation, default_type):
        return dict(type=default_type)

    @classmethod
    def _modify_yaml_doc(cls, yaml_doc, kafe_object_class, **kwargs):
        if kwargs:
            raise ValueError('Received unexpected kwargs: %s' % kwargs)
        return yaml_doc

    @classmethod
    def _get_required_keywords(cls, yaml_doc, kafe_object_class):
        return []

    @classmethod
    def _get_ignored_if_none_keywords(cls):
        return []

    @classmethod
    def _convert_yaml_doc_to_object(cls, yaml_doc):
        return None, None  # format: return <kafe2 object>, <leftover yaml doc>

    # TODO integrate into _modify_yaml_doc
    @classmethod
    def _get_subspace_override_dict(cls, kafe_object_class):
        return dict()

    def read(self):
        with self._ihandle as _h:
            self._yaml_doc = yaml.load(_h, self.LOADER)
        return self._make_object(self._yaml_doc)



r"""This submodule provides utility functions for other modules.

:synopsis: This submodule provides utility functions for other modules.

.. moduleauthor:: Johannes Gaessler <johannes.gaessler@student.kit.edu>
"""

import warnings
from collections import OrderedDict
from typing import List, Optional

import numpy as np

from . import function_library
from .wrapper import *

__all__ = ['string_join_if', 'add_in_quadrature', 'invert_matrix', 'cholesky_decomposition',
           'log_determinant', 'log_determinant_pointwise', 'collect', 'is_diagonal']

# -- general utility functions

def string_join_if(pieces, delim='_', condition=lambda x: x):
    '''Join all elements of `pieces` that pass `condition` together
    using delimiter `delim`.'''
    return delim.join((p for p in pieces if condition(p)))


# -- array/matrix utility functions

def add_in_quadrature(*args):
    '''return the square root of the sum of squares of all arguments'''
    return np.sqrt(np.sum([_a**2 for _a in args], axis=0))


def invert_matrix(mat):
    '''perform matrix inversion'''
    try:
        return np.linalg.inv(mat)
    except np.linalg.LinAlgError:
        warnings.warn(
            "Singular covariance matrix. Are the errors for some data points equal to zero?")
        return None


def cholesky_decomposition(mat):
    """
    Perform Cholesky decomposition of a matrix. Covariance matrices != 0 are always symmetric and
    positive-definite.
    """
    try:
        return np.linalg.cholesky(mat)
    except np.linalg.LinAlgError:
        warnings.warn(
            "Singular covariance matrix. Are the errors for some data points equal to zero?")
        return None


def log_determinant(cholesky_mat):
    """
    Calculate the logarithm of the determinant of a matrix from its Cholesky decomposition.
    """
    if cholesky_mat is None:
        return 0.0  # Easier to handle for multifits than returning None
    else:
        return 2.0 * np.sum(np.log(np.diag(cholesky_mat)))


def log_determinant_pointwise(pointwise_error):
    """
    Calculate the logarithm of the determinant of a matrix from its Cholesky decomposition.
    """
    return 2.0 * np.sum(np.log(pointwise_error))


def collect(*args):
    '''collect arguments into array'''
    return np.asarray(args)


def is_diagonal(matrix):
    """Return True if only diagonal matrix elements are non-zero."""
    return np.all(matrix - np.diag(np.diagonal(matrix)) == 0)

def to_python_types(yaml_dict: Optional[dict]):
    if yaml_dict is None:
        return None
    _new_dict = {}
    for _key, _value in yaml_dict.items():
        if isinstance(_value, float):
            _value = float(_value)  # Casts NumPy scalars to Python float.
        elif isinstance(_value, np.int64):
            _value = int(_value)
        elif isinstance(_value, np.ndarray):
            _value = _value.tolist()
        elif isinstance(_value, OrderedDict):
            _value = [
                float(_v) if not isinstance(_v, np.ndarray) else _v.tolist()
                for _v in _value.values()
            ]
        _new_dict[_key] = _value
    return _new_dict

def to_numpy_arrays(yaml_dict: Optional[dict]):
    if yaml_dict is None:
        return None
    _new_dict = {}
    for _key, _value in yaml_dict.items():
        if isinstance(_value, list):
            _value = np.array(_value)
        _new_dict[_key] = _value
    return _new_dict



"""This submodule provides the necessary objects for parameter estimation from histograms. Currently a histogram needs
to be filled with all individual data points. A function for setting the bin heights is available but not recommended,
as saving and loading those to and from a file is not yet supported.

:synopsis: This submodule provides the necessary objects for parameter estimation from histograms.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .container import *
from .cost import *
from .fit import *
from .model import *
from .plot import *
